/*
 	File:		FontHandlerPrivate.h
 
 	Contains:	QuickDraw GX to PostScript conversion code.
 						File contains private data structures and definitions 
						for the font handler
 
 	Version:	Technology:	Quickdraw GX 1.1.x
 	 
 	Copyright:	 1991-1997 by Apple Computer, Inc., all rights reserved.
*/

#include <Collections.h>
#include "GXPrintingUniverse.h"
#include "FontHandler.h"
#include "FontDatabase.h"
#include "RDUtil.h"



/** This tag goes on a style to indicate what printer font index the style references **/
#define prFontIndexTag 'fhfi'

/** This is the overhead for just reencoding a font. **/
#define kEncodingOverhead 2048

/** Constant for font name index when we couldn't find a font name **/
#define kNoSuchFontName -1

/** Constant for error, no post table 4 **/
#define kNoPOSTtable4		-9999


/** Ultimately, this value should go into "font types.h" **/
#define gxReservedGlyphCode (unsigned short)0xFFFF

/** Use this as a sentinal value to mark unused hash table entries **/
#define fhSentinalGlyphCode gxReservedGlyphCode

/* Indicates Stream Requirements */
enum {

	fhNormalStreaming	=		0,	// Indicates that the font can be streamed normally.
	fhNeedsSnapShots = 		1,	// Indicates that the font can be streamed, but requires variation snapshots.
	fhCantStreamFont = 		2		// Indicates that the font can't be streamed at all.
};

typedef long TFHStreamRequirements;

/***
	Use this value to tag a style to use the two byte font group rather than a particular printer font
	513 is good because you really could never have more than 256 printer fonts so 2*n + 1 seemed good.
	(Actually you could have slightly more than 256 because of Seac, I could figure out the real number but why?)
***/
#define kTwoByteGroupIndex 513


/***************************************
	Some notes on the prFontInfo flags:
	
	fontIsEncoded is the highest level:
	  Only this flag being set means the
		font is ready for use.  A font can
		be on the printer but not encoded
		which means an encoding must be downloaded
		for it.
****************************************/

enum {

	fontNotAvailable = 			0x00000001,		// Font is not available in the printer in any form.
	fontIsInPrinter = 			0x00000002,		// Indicates that font is in printer as resident font.
	fontIsDownloaded =			0x00000004,		// Indicates that font has been downloaded.
	fontIsEncoded = 				0x00000008,		// Indicates that font is encoded and ready for use.
	fontIsPerm = 						0x00000010,		// Indicates that font is downloaded permanently.
	fontCantBeRencoded =		0x00000020,		// Font is a resident composite font, ignore hash table, use mapping from Post Table 4.
	fontIsEncodingOnly = 		0x00000040,		// Font is only a re-encoding of another font.
	fontHasAllDocGlyphs=		0x00000080,		// Download all document glyphs for this printer font except those thar are printer-resident.
	fontIsInSaveLevel =			0x00000100,		// Font was downloaded witin a save-level.
	fontIsTwoByteMember = 	0x00000200,		// Printer Font is member of a synthetic 2 byte font created by Font Handler.
	fontCantBeUsed =				0x80000000		// Font can't be used on printer, client must take other measures to draw shape (such as rasterize on host)
};

typedef long prFontInfo;

typedef struct {

	fhFont						theFont;									// Snapshot Font reference for this printer font.
	gxFont						theRealFont;							// The actual font reference.
	
	long							dbIndex;									// The glyph/bit array index in the font database (which snapshot for the font)
	
	long							prFontIndex;							// The printer font index for this printer font.
	long							docVMUsage;								// How much does this printer-font's document font need. 
																							//	(the same for all printer-fonts for a given doc font)
	long							vmUsage;									// How much VM does printer font use.
	prFontInfo				info;											// Information about the font.
	long							nameIndex;								// Cache name index here.
	unsigned short		printerFontCodes[256];		// Hash table array of printer font codes.
	Handle						glyphToCharCodeMap;				// Mapping for glyph codes to character codes used instead of searching hash table.
	long							usageSize;								// size of usage array (long alligned)
	unsigned long			glyphUsage[1];						// array of longs for glyph usage (range or bits)

} TPrinterFontRec;

#define kMaxWrkSpaces 6

typedef struct {

	CGXtoPostScriptDevice	*psDevice;							// Object for device interaction.
	scalerStreamTypeFlag	legalStreamTypes;				// legal types for font streaming.
	short									asciiDownload;					// Ascii download modifier for streaming necessary?
	Boolean								useTypeZeroFonts;				// True if printer can support type-0 (composite) fonts for 2 byte.
	Boolean								printerCanDoAllFonts;		// False if the printer cannot handle certain fonts (2-byte on some printers for example).
	char*									productDescription;			// C string describing PostScript product. (version revision product-name)
	long									printerVM;							// How much vm availabe on printer for fonts.
	long									vmAvailable;						// How much is left.
	long									saveLevel;							// boolean indicates page-level save was done.
	
	scalerStreamTypeFlag	documentStreamTypes;		// Stream types used by document based on initial database.

	Handle								workHandle[kMaxWrkSpaces];			// Workspace handles for miscellanious stuff.
	long									workSize[kMaxWrkSpaces];				// Size of workspace handles.
#if DEBUGLEVEL > 0
	Boolean								workHandleInUse[kMaxWrkSpaces];	// Handle in use?
#endif

	TRDMapHdl							rdMap;									// RDUtils map.
	TFontDbase						docDbase;								// document font data base.
	Collection						snapShots;							// Collection of snapshot records.
	long									nextSnapShot;						// Next id to use for a snapshot font.
	Collection						printerFontsOffsets;		// Collection of offsets into printer font list.
	long									numPrFonts;							// Number of printer fonts there are.
	long									nextOffset;							// next offset into printer font data.
	unsigned char					printerFonts[1];				// packed array of Printer font data (of type TPrinterFontRec)
	
} TFontHandlerRec, *TFontHandlerPtr, **TFontHandlerHdl;


/* Deal with Intrasegment calling of published routines */


/************************* Private routines **************************/

/** Routines for memory management **/

unsigned short FHGetGlyphHash(long glyphIndex, unsigned short *printerCodes);

OSErr	FHSetWorkHandleSize(TFontHandlerHdl hFHRec, long size, long index, Handle *h);

void FHReleaseWorkspace(TFontHandlerHdl hFHRec, long index);


/** Other stuff **/

OSErr FHGetStreamRequirements(gxFont theFont, CGXtoPostScriptDevice *psDevice, scalerStreamTypeFlag legalStreamTypes, 
															char *productDescription, TFHStreamRequirements *requirements);

long FHGeneratePrinterName(TPrinterFontRec *printerFont, unsigned char *name);

long FHGetStylePrFontIndex(gxStyle theStyle);

long	CountGlyphBits(unsigned long* glyphBits, long nGlyphs);

OSErr FHAddPrinterFonts(TFontHandlerHdl hFHRec, fhFont theFont, unsigned long *glyphBits, long nGlyphs, TfdbInfoFlags fdbFlags);

OSErr FHFindDocLevelFonts(TFontHandlerHdl hFHRec);

OSErr FHComputePrinterFontSize(TFontHandlerHdl hFHRec, TPrinterFontRec *printerFont);

OSErr	FHComputeDocFontSize(TFontHandlerHdl hFHRec, fhFont theFont, long nGlyphs, 
														unsigned long *glyphBits,	unsigned long *residentBits, long *vmSize);

OSErr FHDownloadFont(TFontHandlerPtr pFHRec, TPrinterFontRec *printerFont, scalerStreamTypeFlag legalTypes, char *productDescription);

OSErr FHReEncodeFont(TFontHandlerPtr pFHRec, TPrinterFontRec *printerFont, TPrinterFontRec *printerFontToEncode);

OSErr FHAddNonDatabaseShape(TFontHandlerHdl hFHRec, gxShape theShape);

gxFont FHgxGetStyleFont(gxStyle theStyle);

OSErr FHTestFontCanBeUsed(TFontHandlerHdl hFHRec, fhFont theFont, long glyphsUsed, long glyphsEncoded, Boolean *canUse);

OSErr	FHCanPrinterHandleAllFonts(unsigned char *productName, Boolean *printerCan);

Boolean FHFontMayBe2Byte(gxFont theFont);

/** Routines for accessing printer font records **/

OSErr FHGetIndexedPrinterFont(TFontHandlerHdl hFHRec, fhFont theFont, long index, TPrinterFontRec* *printerFont, long* theOffset);

long		FHCountPrinterFonts(TFontHandlerHdl hFHRec, fhFont theFont);



/** Routines for dealing with synthetic composite fonts **/

long FHGenerateTwoByteFontName(TFontHandlerHdl hFHRec, fhFont theFont, unsigned char name[256]);

OSErr FHMakeTwoByteFont(TFontHandlerHdl hFHRec, fhFont theFont);


/** Utility routines **/

long FHGetFontName(gxFont fontID, long index, gxFontName *name, gxFontPlatform *platform,
		gxFontScript *script, gxFontLanguage *language, unsigned char text[]);

