/*
 	File:		GXtoPSRenderTmodes.cp
 
 	Contains:	QuickDraw GX to PostScript conversion code.
 
 	Version:	Technology:	Quickdraw GX 1.1.x
 	 
 	Copyright:	 1994-1997 by Apple Computer, Inc., all rights reserved.
*/

/************************
	
	GXtoPSRenderTmodes.cp

	This file contains the code necessary to render the 
	transfer modes as a full color image at device resolution.
	
	Note, there are some resolution limitations here since a 
	GX rectangle and bitmap can only be 32k wide.  
	
	This means we cannot do more then about 10" on a 3000dpi
	device.  This wouldn't be practical anyway.  Probably
	better to use the geometric method for high resolution devices.
	
	Since we'd like to inherit all of the wonderful capabilities of the banding
	and color matching built into the BitmapPrimitive of the imaging engine,
	
	We're going to make a special bitmap shape here, only we'll communicate to the 
	Imaging engine what we are really doing a "special" bitmap explained below.

*************************/

#include "GXToPSBuildConfig.h"
#include "GXToPostScript.h"
#include <GXErrors.h>
#include "Private.h"			// we need to get at some Imaging Engine routines.
#include "PublicPostScriptIE.h"


/*************

	This function rerenders the shape into a device resolution image in the device's color space.
	This way we get transfer modes.
	
	theShape:				The picture shape for the page.
	tModeShape:			A picture (or single shape) that represents only the shapes
									in the page (theShape) that contain transfer modes.  Typically,
									this will have been accumulated during the normal GX->PS translation
									phase of the DrawShape process.

**************/
void CGXtoPostScriptTranslator::RenderTransferModes(gxShape theShape, gxShape tModeShape)
{
	/* First find out the bounding box of the area of theShape that had transfer modes.
			This will be the bounds of tModeShape.  
	*/
	gxRectangle			tModeBounds;
	GXGetShapeLocalBounds(tModeShape, &tModeBounds);

	/* Now find out the bounds in device resolution */
	gxPoint					topLeft, bottomRight;	
	topLeft.x = tModeBounds.left;
	topLeft.y = tModeBounds.top;
	bottomRight.x = tModeBounds.right;
	bottomRight.	y = tModeBounds.bottom;
	MapPoints(&mDeviceMapping, 1, &topLeft);
	MapPoints(&mDeviceMapping, 1, &bottomRight);

	/* Now find out the bounds of the image for the transfer modes */
	long width = FixedRound(bottomRight.x) - FixedRound(topLeft.x);
	long height = FixedRound(bottomRight.y) - FixedRound(topLeft.y);
	
	/*	
		Make the special "Bitmap" shape that the imaging engine can draw.
		This is a bitmap shape where we'll use a tag to indicate to the imaging
		engine that the data do band is really the picture of the page.
		
		Our special bitmap is:
			image field points to the address of the shape to draw
			rowBytes is -sizeof(gxShape)
	*/
	gxBitmap			theBits;
	theBits.width = width;
	theBits.height = height;
	theBits.profile = mDevProfile;
	theBits.set = nil;
	theBits.image = (char*)theShape;
	theBits.rowBytes = -sizeof(gxShape);

	switch (mDevColorSpace) {
		case gxRGBSpace:
			theBits.space = gxRGB32Space;
			theBits.pixelSize = 32;
			break;
		case gxCMYKSpace:
			theBits.space = gxCMYK32Space;
			theBits.pixelSize = 32;
			break;
		case gxGraySpace:
			theBits.space = gxIndexedSpace;
			theBits.set = CreateGrayColorSet();			
			theBits.pixelSize = 8;
			break;
		default:
			throw colorSpace_out_of_range;
			break;
	}//end switch

	/* Make the "bitmap shape, with a transform that will scale it down to the user space */
	gxShape			specialBits = GXNewBitmap(&theBits, &topLeft);
	gxTransform	specialTransform = GXNewTransform();
	gxMapping		specialMapping;
	InvertMapping(&specialMapping, &mDeviceMapping);
	GXSetTransformMapping(specialTransform, &specialMapping);
	GXSetShapeTransform(specialBits, specialTransform);
	GXDisposeTransform(specialTransform);												// shape owns it now, we don't need it.
	if (theBits.set != nil) GXDisposeColorSet(theBits.set);			// shape owns it now, we don't need it.
	
	/* Now "draw" the "special" bitmap */
	OSErr status = PostScriptDrawShape(mPSIEContext, specialBits, nil, 0);
	if (status != noErr)
		throw status;
	
	/* Now clean up. */
	
	GXDisposeShape(specialBits);	

	/* For debugging purposes, print the transfer-mode bounding box */
	#if DRAWTMODEBOUNDS
	{
		gxShape			tmBoundsShape = GXNewRectangle(&tModeBounds);
		GXSetShapeFill(tmBoundsShape, gxClosedFrameFill);
		status = PostScriptDrawShape(mPSIEContext, tmBoundsShape, nil, 0);
		GXDisposeShape(tmBoundsShape);
		if (status != noErr)
			throw status;
	}
	#endif
	
}//CGXtoPostScriptDevice::RenderTransferModes

